#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Virginia report views.
"""

from urllib import urlencode
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.traversing.browser.absoluteurl import absoluteURL
from zc.table.column import GetterColumn
import zc.resourcelibrary

from schooltool.basicperson.interfaces import IDemographics
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import ISectionContainer
from schooltool.cando.browser import report
from schooltool.cando.interfaces import ISectionSkills
from schooltool.cando.interfaces import ISkillsGradebook
from schooltool.cando.interfaces import IStudentIEP
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish
from schooltool import table

from schooltool.virginia import VirginiaMessage as _


class VASectionReportTablesViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile('templates/va_section_report_tables.pt')

    @property
    def passing_scores_row(self):
        result = []
        for passing_score in self.view.passing_scores:
            if passing_score['value'] >= self.view.passing_score:
                result.append(passing_score)
        return '/'.join(reversed([i['submit_value'] for i in result]))


class VASectionReportPassingScoreViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile(
        'templates/va_section_report_passing_score.pt')

    def options(self):
        result = []
        for passing_score in self.view.passing_scores:
            selected = passing_score['value'] == self.view.passing_score
            result.append({
                    'value': passing_score['submit_value'],
                    'title': passing_score['label'],
                    'selected': selected,
                    })
        return result


class VAReportViewBase(object):

    def render(self, *args, **kw):
        zc.resourcelibrary.need('schooltool.virginia.flourish')
        return super(VAReportViewBase, self).render(*args, **kw)


class VASectionReportView(VAReportViewBase,
                          report.SectionReportViewBase):

    container_class = 'container extra-wide-container'

    def update(self):
        rows = []
        collator = ICollator(self.request.locale)
        worksheets = self.context.__parent__.__parent__
        students_count = 0
        for student in sorted(self.section.members,
                              key=lambda x:x.title,
                              cmp=collator.cmp):
            demographics = IDemographics(student)
            info = {
                'student': student.title,
                'student_id': student.username,
                'required_container_id': 'required-%s' % student.username,
                'all_container_id': 'all-%s' % student.username,
                'student_id': demographics.get('ID', ''),
                }
            info.update(self.getStudentSkillsData(student, worksheets))
            rows.append(info)
        self.rows = rows
        attainment_labels = {100: '90-100%',
                             90: '80-89%',
                             80: '70-79%',
                             70: '60-69%',
                             60: '<= 59%'}
        rows_attainment = {100: 0,
                           90: 0,
                           80: 0,
                           70: 0,
                           60: 0}
        rows_attainment80 = {u'80% or above': 0,
                             u'Below 80%': 0}
        for row in self.rows:
            if row['all_evaluated']:
                students_count += 1
                rows_attainment[row['rows_attainment']] += 1
                rows_attainment80[row['rows_attainment80']] += 1
        self.rows_attainment = []
        for key in sorted(rows_attainment, reverse=True):
            info = {"scale": attainment_labels[key],
                    "student_count": rows_attainment[key]}
            try:
                p = rows_attainment[key] / float(students_count)
            except (ZeroDivisionError,):
                p = 0
            info["percentage"] = self.formatPercentage(p)
            self.rows_attainment.append(info)
        self.rows_attainment80 = []
        for key in sorted(rows_attainment80):
            info = {"scale": key,
                    "student_count": rows_attainment80[key]}
            try:
                p = rows_attainment80[key] / float(students_count)
            except (ZeroDivisionError,):
                p = 0
            info["percentage"] = self.formatPercentage(p)
            self.rows_attainment80.append(info)

    def getStudentSkillsData(self, student, worksheets):
        required_total = 0
        required_evaluated = 0
        required_competent = 0
        all_total = 0
        all_evaluated = 0
        all_competent = 0
        for worksheet in worksheets.values():
            gradebook = ISkillsGradebook(worksheet)
            for skill in gradebook.activities:
                all_total += 1
                score = gradebook.getScore(student, skill)
                evaluated = score is not None and score.value is not UNSCORED
                # XXX: use real scoresystem values here instead of int()
                competent = evaluated and \
                            int(score.value) >= int(self.passing_score)
                if evaluated:
                    all_evaluated += 1
                if competent:
                    all_competent += 1
                if skill.required:
                    required_total += 1
                    if evaluated:
                        required_evaluated += 1
                    if competent:
                        required_competent += 1
        try:
            required_rating = required_competent / float(required_evaluated)
        except (ZeroDivisionError,):
            required_rating = 0
        try:
            required_completed = required_competent / float(required_total)
        except (ZeroDivisionError,):
            required_completed = 0
        try:
            all_rating = all_competent / float(all_evaluated)
        except (ZeroDivisionError,):
            all_rating = 0
        required_completed_percentage = required_rating * 100
        if required_completed_percentage >= 90:
            rows_attainment = 100
        elif required_completed_percentage >= 80:
            rows_attainment = 90
        elif required_completed_percentage >= 70:
            rows_attainment = 80
        elif required_completed_percentage >= 60:
            rows_attainment = 70
        else:
            rows_attainment = 60
        if required_completed_percentage >= 80:
            rows_attainment80 = u'80% or above'
        else:
            rows_attainment80 = u'Below 80%'
        return {
            'required_total': required_total,
            'required_evaluated': required_evaluated,
            'required_competent': required_competent,
            'required_rating': self.formatPercentage(required_rating),
            'required_completed': self.formatPercentage(required_completed),
            'required_data': [required_total, required_evaluated, required_competent],
            'all_total': all_total,
            'all_evaluated': all_evaluated,
            'all_competent': all_competent,
            'all_rating': self.formatPercentage(all_rating),
            'all_data': [all_total, all_evaluated, all_competent],
            'rows_attainment': rows_attainment,
            'rows_attainment80': rows_attainment80,
            }

    def formatPercentage(self, number):
        return '%.0f%%' % (number * 100)


class VAReportDetailsViewlet(report.SectionReportDetailsViewletBase):

    template = ViewPageTemplateFile(
        'templates/va_section_report_details.pt')


class VASkillsCompletionReportView(VAReportViewBase,
                                   report.SkillsCompletionReportViewBase):

    default_skill_type = 'required'
    container_class = 'container extra-wide-container'

    def isIEPSkill(self, iep_skills, skill):
        skillset = skill.__parent__
        return skillset in iep_skills and skill in iep_skills[skillset]

    def sectionSkillsData(self, section):
        result = {}
        students_count = len(list(section.members))
        section_skills = {}
        for worksheet in ISectionSkills(section).values():
            gradebook = ISkillsGradebook(worksheet)
            for skill in worksheet.values():
                section_skills[skill] = gradebook
        students_evaluated = []
        section_required_skills = []
        section_skills_evaluated = []
        section_skills_passed = []
        section_attainment_of_required = []
        section_attainment_of_evaluated = []
        for student in section.members:
            iep = IStudentIEP(student)
            iep_skills = iep.getIEPSkills(section)
            required = []
            evaluated = []
            passed = []
            attainment_of_required = 0.0
            attainment_of_evaluated = 0.0
            for skill, gradebook in section_skills.items():
                if skill.required and not self.isIEPSkill(iep_skills, skill):
                    required.append(1)
                    score = gradebook.getScore(student, skill)
                    if score is not None and score.value is not UNSCORED:
                        evaluated.append(1)
                        value = int(score.value)
                        if value >= self.passing_score:
                            passed.append(1)
            if required:
                section_required_skills.append(len(required))
                attainment_of_required = len(passed) / float(len(required))
                section_attainment_of_required.append(attainment_of_required)
                if evaluated:
                    students_evaluated.append(1)
                    section_skills_evaluated.append(len(evaluated))
                    section_skills_passed.append(len(passed))
                    attainment_of_evaluated = len(passed) / float(len(evaluated))
                    section_attainment_of_evaluated.append(attainment_of_evaluated)
        result['students_count'] = float(students_count)
        result['students_evaluated'] = float(len(students_evaluated))
        result['total_skills'] = float(sum(section_required_skills))
        result['skills_evaluated'] = float(sum(section_skills_evaluated))
        result['skills_passed'] = float(sum(section_skills_passed))

        if result['students_evaluated']:
            result['section_required_skills'] = round(result['total_skills'] / result['students_evaluated'])
            result['completion_3'] = len([x for x in section_attainment_of_evaluated if x >= 0.8]) / result['students_evaluated']
            result['progress_2'] = result['skills_evaluated'] / result['students_evaluated']
            result['progress_3'] = result['skills_passed'] / result['students_evaluated']
        else:
            result['section_required_skills'] = 0
            result['completion_3'] = 0
            result['progress_2'] = 0
            result['progress_3'] = 0

        # completion report specific columns
        result['completion_1'] = result['students_count']
        result['completion_2'] = result['section_required_skills']

        # progress report specific columns
        result['progress_1'] = result['section_required_skills']
        if result['skills_evaluated']:
            result['progress_4'] = result['skills_passed'] / result['skills_evaluated']
        else:
            result['progress_4'] = 0
        if result['total_skills']:
            result['progress_5'] = result['skills_passed'] / result['total_skills']
        else:
            result['progress_5'] = 0
        return result

    def getSectionData(self, section):
        result = {
            'section': section,
            'container_id': 'skills-data-%s' % section.__name__,
            'courses': set(list(section.courses)),
            }
        result.update(self.sectionSkillsData(section))
        return result


class VASkillsAttainmentReportView(VASkillsCompletionReportView):

    pass


class VASkillsCompletionReportTableFilter(table.ajax.TableFilter):

    template = ViewPageTemplateFile(
        'templates/va_completion_report_table_filter.pt')

    def passing_score_options(self):
        result = []
        for passing_score in self.view.passing_scores:
            selected = passing_score['value'] == self.view.passing_score
            result.append({
                    'value': passing_score['submit_value'],
                    'title': passing_score['label'],
                    'selected': selected,
                    })
        return result

    def location_options(self):
        result = []
        for location in self.view.locations:
            selected = location['value'] == self.view.location
            result.append({
                    'value': location['submit_value'],
                    'title': location['label'],
                    'selected': selected,
                    })
        return result

    def course_options(self):
        result = []
        for course in self.view.courses:
            selected = course['value'] == self.view.course
            result.append({
                    'value': course['submit_value'],
                    'title': course['label'],
                    'selected': selected,
                    })
        return result

    def filter(self, items):
        if not self.manager.fromPublication:
            return []
        course = self.view.course
        if course is not None:
            items = filter(lambda x: course in x.courses, items)
        location = self.view.location
        if location is not None:
            items = filter(lambda x: location in x.resources, items)
        result = [self.view.getSectionData(section) for section in items]
        if result:
            totals = self.getTotals(result)
            result.append(totals)
        return result

    def getTotals(self, sections_data):
        result = {
            'instructors': None,
            'course': None,
            'section': report.placeholder,
            }
        completion_1 = []
        completion_2 = []
        completion_3 = []
        progress_1 = []
        progress_2 = []
        progress_3 = []
        progress_4 = []
        progress_5 = []
        for data in sections_data:
            completion_1.append(data['completion_1'])
            completion_2.append(data['completion_2'])
            completion_3.append(data['completion_3'])
            progress_1.append(data['progress_1'])
            progress_2.append(data['progress_2'])
            progress_3.append(data['progress_3'])
            progress_4.append(data['progress_4'])
            progress_5.append(data['progress_5'])
        result['completion_1'] = sum(completion_1)
        result['completion_2'] = sum(completion_2)
        result['completion_3'] = sum(completion_3) / len(sections_data)
        result['progress_1'] = sum(progress_1)
        result['progress_2'] = sum(progress_2) / len(sections_data)
        result['progress_3'] = sum(progress_3) / len(sections_data)
        result['progress_4'] = sum(progress_4) / len(sections_data)
        result['progress_5'] = sum(progress_5) / len(sections_data)
        return result


class VATotalsSortTableFormatter(table.ajax.AJAXFormSortFormatter):

    @property
    def fixed_sort_on(self):
        return (('totals', False),)


class VASkillsTermReportTable(table.ajax.Table):

    batch_size = 0
    table_formatter = VATotalsSortTableFormatter
    css_classes = ''

    @property
    def html_id(self):
        return self.view.form_container_id
 
    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': self.css_classes})

    def items(self):
        return ISectionContainer(self.context).values()

    def sortOn(self):
        return (
            ('totals', False),
            ('instructors', False),
            ('course', False),
            ('section', False)
            )


class TotalsColumn(GetterColumn):

    def getSortKey(self, item, formatter):
        return item['section'] is report.placeholder


class VASkillsCompletionReportTable(VASkillsTermReportTable):

    css_classes = 'data section-report va-skills-report va-skills-completion-report'
    visible_column_names = [
        'instructors',
        'course',
        'section',
        'progress_1',
        'progress_2',
        'progress_3',
        'progress_4',
        'progress_5',
        ]

    def columns(self):
        instructors = table.column.LocaleAwareGetterColumn(
            name='instructors',
            title=_('Teachers'),
            getter=report.get_section_instructors,
            cell_formatter=report.section_instructors_formatter,
            subsort=True)
        course = table.column.LocaleAwareGetterColumn(
            name='course',
            title=_('Course'),
            getter=report.get_section_courses,
            subsort=True)
        section = table.column.LocaleAwareGetterColumn(
            name='section',
            title=_('Section'),
            getter=report.get_section_title,
            subsort=True)
        required_skills_count = GetterColumn(
            name='progress_1',
            title=_('# Required Skills'),
            getter=lambda i, f: '%0.f' % i['progress_1'])
        average_evaluated_skills_count = GetterColumn(
            name='progress_2',
            title=_('Avg. # Skills Evaluated '),
            getter=lambda i, f: '%0.f' % i['progress_2'])
        average_evaluated_skills_attained_count = GetterColumn(
            name='progress_3',
            title=_('Avg. # Evaluated Skills Attained'),
            getter=lambda i, f: '%0.f' % i['progress_3'])
        average_evaluated_skills_attained_percentage = GetterColumn(
            name='progress_4',
            title=_('Avg. % Evaluated Skills Attained'),
            getter=lambda i, f: '%0.f' % (i['progress_4']*100))
        average_required_skills_attained_percentage = GetterColumn(
            name='progress_5',
            title=_('Avg. % Required Skills Attained'),
            getter=lambda i, f: '%0.f' % (i['progress_5']*100))
        return [
            TotalsColumn(name='totals', subsort=True),
            instructors,
            course,
            section,
            required_skills_count,
            average_evaluated_skills_count,
            average_evaluated_skills_attained_count,
            average_evaluated_skills_attained_percentage,
            average_required_skills_attained_percentage,
            ]


class VASkillsAttainmentReportTable(VASkillsTermReportTable):

    css_classes = 'data section-report va-skills-report va-skills-attainment-report'
    visible_column_names = [
        'instructors',
        'course',
        'section',
        'completion_1',
        'completion_2',
        'completion_3',
        ]

    def columns(self):
        instructors = table.column.LocaleAwareGetterColumn(
            name='instructors',
            title=_('Teachers'),
            getter=report.get_section_instructors,
            cell_formatter=report.section_instructors_formatter,
            subsort=True)
        course = table.column.LocaleAwareGetterColumn(
            name='course',
            title=_('Course'),
            getter=report.get_section_courses,
            subsort=True)
        section = table.column.LocaleAwareGetterColumn(
            name='section',
            title=_('Section'),
            getter=report.get_section_title,
            subsort=True)
        students_count = GetterColumn(
            name='completion_1',
            title=_('# of Students'),
            getter=lambda i, f: '%0.f' % i['completion_1'])
        required_skills_count = GetterColumn(
            name='completion_2',
            title=_('# of Skills'),
            getter=lambda i, f: '%0.f' % i['completion_2'])
        students_above_80_percent_attainment = GetterColumn(
            name='completion_3',
            title=_('% of Students >= 80% Attainment'),
            getter=lambda i, f: '%0.f' % (i['completion_3']*100))
        return [
            TotalsColumn(name='totals', subsort=True),
            instructors,
            course,
            section,
            students_count,
            required_skills_count,
            students_above_80_percent_attainment,
            ]


class ExportAsPDFLinkViewlet(flourish.viewlet.Viewlet):

    html_id = 'export-as-pdf-link'
    container_id = 'export-as-pdf-container'
    template = ViewPageTemplateFile('templates/export_as_pdf_link.pt')

    def url(self):
        passing_score_filter_id = self.view.passing_score_filter_id
        location_filter_id = self.view.location_filter_id
        course_filter_id = self.view.course_filter_id
        params = {
            passing_score_filter_id: self.request.get(passing_score_filter_id),
            location_filter_id: self.request.get(location_filter_id),
            course_filter_id: self.request.get(course_filter_id),
            'file_type': 'PDF',
            'description': _('Virginia Required Skills Completion Report'),
            }
        return '%s/%s?%s' % (absoluteURL(self.context, self.request),
                             'request_va_skills_attainment_report.html',
                             urlencode(params))

    def render(self):
        if self.manager.fromPublication and self.manager.batch.length:
            return self.template(self)
        return ''

    @property
    def dialog_title(self):
        return translate(_('Virginia Required Skills Completion Report'),
                         context=self.request)


class StickyTableHeaderViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <script>
        $('table').stickyTableHeaders();
      </script>
    ''')

    def render(self):
        if self.manager.fromPublication and self.manager.batch.length:
            return self.template(self)
        return ''


class VASkillsAttainmentReportPDFView(flourish.report.PlainPDFPage,
                                      VASkillsAttainmentReportView):

    name = _('Virginia Required Skills Completion Report')

    @property
    def title(self):
        return self.context.title

    @property
    def scope(self):
        term = self.context
        schoolyear = ISchoolYear(term)
        return '%s | %s' % (term.title, schoolyear.title)

    @property
    def subtitles_left(self):
        result = [_('Passing Score: ${passing_score}',
                    mapping={'passing_score': self.passing_score})]
        if self.course is not None:
            result.append(
                _('Course: ${course}', mapping={'course': self.course.title}))
        else:
            result.append(_('Course: All'))
        if self.location is not None:
            result.append(
                _('Location: ${location}',
                  mapping={'location': self.location.title}))
        else:
            result.append(_('Location: All'))
        return result

    @property
    def base_filename(self):
        term = self.context
        schoolyear = ISchoolYear(term)
        return 'va_skills_completion_report_%s_%s' % (
            schoolyear.__name__,
            term.__name__)


class RequestVASkillsAttainmentReportView(RequestRemoteReportDialog,
                                          VASkillsCompletionReportView):

    report_builder = VASkillsAttainmentReportPDFView

    template = ViewPageTemplateFile(
        'templates/request_va_skills_attainment_report.pt')

    @property
    def params(self):
        params = [
            self.passing_score_filter_id,
            self.location_filter_id,
            self.course_filter_id,
            'file_type',
            'description',
            ]
        result = []
        for param in params:
            result.append({
                    'name': param,
                    'value': self.request.get(param),
                    })
        return result

    def schedule(self, task):
        params = {}
        for param in self.params:
            params[param['name']] = param['value']
        task.request_params.update(params)
        task.schedule(self.request)

        
class VASkillsAttainmentReportPDFViewPart(table.pdf.RMLTablePart):

    template = flourish.templates.XMLFile('rml/paragraph_table_headers.pt')

    table_name = 'skills_completion_report_table'

    def getColumnWidths(self, rml_columns):
        return '20% 25% 25% 8% 7% 15%'


class VASkillsAttainmentReportPDFViewTableFilter(
    VASkillsCompletionReportTableFilter):

    def filter(self, items):
        location = self.view.location
        course = self.view.course
        if course is not None:
            items = filter(lambda x: course in x.courses, items)
        if location is not None:
            items = filter(lambda x: location in x.resources, items)
        result = [self.view.getSectionData(section) for section in items]
        if result:
            totals = self.getTotals(result)
            result.append(totals)
        return result

    def render(self, *args, **kw):
        return ''
